'use strict';

HLL('module:headland.postgres.Row.js', function () {
	HL.loaded(['headland', 'postgres', 'Row'], CL.fun(function* (CL, HL, $$module) {
		let Model = yield HL.load(['headland', 'Model'], false, false);

		let _2;

		let hop = Object.prototype.hasOwnProperty;

		let ops = {
			'=': '=',
			'<': '<',
			'>': '>',
			'<=': '<=',
			'>=': '>=',
			'!=': '!=',
			'&&': '&&',
			'@@': '@@'
		};

		let fkey = function (ikey) {
			if (ikey instanceof QueryBuilder.Raw) return ikey;
			let key = ikey;
			let match = key.match(/^(.*)_id$/);
			if (match) key = match[1];
			return '"' + key.replace('"', '\\"') + '"';
		};

		let _superInit = Model.prototype.init;
		let Row = Model.extend({
			name: ['headland', 'postgres', 'Row'],
			mode: 'shared',
			fields: {
				init: {
					method: function (data, exists) {
						let _this = this;

						let _private = this['private:headland.postgres.Row'];

						if (exists) {
							let fields = this.static.classInfo.fields;
							for (let key in data) {
								let field = fields[key];
								let type = field && field.type && field.type();
								let val = data[key];
								if (field && field.sub && val) {
									data[key] = field.sub().fromJSON(typeof val == 'object' ? val : JSON.parse(val));
								}
							}
						}
						let secret = _superInit.call(_this, data, 'cluster', true);
						_private.data = secret.data;
						_private.state = secret.state;
						if (!exists) _private.insert();
					}
				},
				saveChanges: {
					method: CL.fun(function* () {
						let _private = this['private:headland.postgres.Row'];

						if (_private.state.saved) return;
						_private.state.saved = true;

						let fields = this.static.classInfo.fields;

						let q = new QueryBuilder('UPDATE ' + fkey(this.static.table) + ' SET');

						let i = 0;
						let changes = _private.state.changes;
						_private.state.changes = {};
						_2.each(changes, function (val, key) {
							let field = fields[key];
							if (!field.type) return;
							if (i++) q.push(',');
							let proper;
							if (typeof val == 'object') {
								let type = field.type();
								if (type.indexOf('json') == 0) proper = JSON.stringify(val);else if (type == 'point') proper = '(' + (val.x || val.lon || val.lng || val[0]) + ',' + (val.y || val.lat || val[1]) + ')';else proper = JSON.parse(JSON.stringify(val));
							} else {
								proper = val;
							}
							q.push([fkey(key), '=', [proper]]);
						});

						q.push(['WHERE', fkey('id'), '=', [_private.data.id]]);

						//console.log(q.get());

						if (i) yield Connection.query(q.get());

						if (_private.state.trigger.length) {
							this.trigger('change', { changes: _2.pick(changes, _private.state.trigger), index: this.getIndexData() });
							_private.state.trigger = [];
						}
					})
				}
			},
			privateFields: {
				data: {},
				state: {},
				insert: {
					method: CL.fun(function* () {
						let _private = this;

						return CL.fun(function* () {
							let q = new QueryBuilder('INSERT INTO ' + fkey(this.static.table));
							let keys = [];
							let values = [];
							let fields = this.static.classInfo.fields;

							_2.each(_private.data, function (val, key) {
								if (val === void 0) return;
								if (keys.length) {
									keys.push(',');
									values.push(',');
								}
								keys.push(fkey(key));
								let field = fields[key];
								let proper;
								if (typeof val == 'object') {
									let type = field.type && field.type();
									if (type && type.indexOf('json') == 0) proper = JSON.stringify(val);else if (type == 'point') proper = '(' + (val.x || val.lon || val.lng || val[0]) + ',' + (val.y || val.lat || val[1]) + ')';else proper = JSON.parse(JSON.stringify(val));
								} else {
									proper = val;
								}
								values.push([proper]);
							});
							q.push('(', keys, ') VALUES (', values, ') RETURNING *');

							//console.log(q.get());

							let res = yield Connection.query(q.get());
							let hadId = _private.data.id !== void 0;
							_2.each(res.rows[0], function (val, key) {
								if (_private.data[key] === void 0) _private.data[key] = val;
							});

							if (!hadId) this.registerObservable('cluster');
							this.trigger('create', { data: _private.data });
						}).apply(this.public);
					})
				}
			},
			statics: {
				fetch: {
					method: 'proxy',
					mode: 'proxy'
				},
				select: {
					method: CL.fun(function* (opts) {
						let _private = this['private:headland.postgres.Row'];


						let q = new QueryBuilder('SELECT');

						if (opts.count || opts.return == 'count') {
							q.push('COUNT(*) AS ' + fkey('count'));
						} else if (opts.return == 'ids') {
							q.push('id');
						} else if (opts.return == 'data' && opts.select) {
							for (let i = 0; i < opts.select.length; i++) {
								if (i) q.push(',');
								q.push(opts.select[i]);
							}
						} else {
							q.push('*');
						}

						q.push('FROM', fkey(this.table));

						if (opts.join && opts.join.length) {
							q.push('JOIN');
							for (let i = 0; i < opts.join.length; i++) {
								let join = opts.join[i];
								if (i) q.push(',');
								if (join instanceof QueryBuilder.Raw) {
									q.push(join);
								} else {
									CL.throw('HLPostgreSQLSelect', 'Proper join not implemented, use Raw');
								}
							}
						}

						if (opts.filter) {
							opts.filters = opts.filters || [];
							opts.filters.push(opts.filter);
						}
						if (opts.filters && opts.filters.length) {
							let added = 0;
							let push = function () {
								q.push(added++ ? 'AND' : 'WHERE');
								q.push.apply(q, arguments);
							};
							for (let filter of opts.filters) {
								if (_2.isArray(filter)) {
									if (ops[filter[1]]) push([fkey(filter[0]), ops[filter[1]], [filter[2]]]);else if (filter[1] == 'null') push(fkey(filter[0]), 'IS NULL');else if (filter[1] == '!null') push(fkey(filter[0]), 'IS NOT NULL');else if (filter[1] == 'contains') push([[filter[2]], '= ANY(' + fkey(filter[0]) + ')']);else if (filter[1] == 'in') push([fkey(filter[0]), '= ANY(', [filter[2]], ')']);else CL.throw('HLPostgreSQLSelect', 'Unknown filter passed to select()');
								} else if (filter instanceof QueryBuilder.Raw) {
									push(filter);
								} else if (typeof filter == 'object') {
									_2.each(filter, function (val, key) {
										push([fkey(key), '=', [val]]);
									});
								} else {
									CL.throw('HLPostgreSQLSelect', 'Unknown filter passed to select()');
								}
							}
						}

						if (opts.group && opts.group.length) {
							q.push('GROUP BY');
							for (let i = 0; i < opts.group.length; i++) {
								if (i) q.push(',');
								q.push(fkey(opts.group[i]));
							}
						}

						if (opts.order && opts.order.length) {
							q.push('ORDER BY');
							for (let i = 0; i < opts.order.length; i++) {
								let key = opts.order[i][0];
								let dir = opts.order[i][1];
								if (i) q.push(',');
								q.push(fkey(key), dir.toUpperCase());
							}
						}

						if (opts.offset) q.push('OFFSET', [[opts.offset]]);
						if (opts.limit) q.push('LIMIT', [[opts.limit]]);

						if (opts.return == 'query') return q;

						if (opts.count || opts.return == 'count') {
							let rq = q instanceof QueryBuilder ? q.get() : q;
							let res = yield Connection.query(rq);
							return res.rows[0].count;
						} else if (opts.return == 'ids') {
							let rq = q instanceof QueryBuilder ? q.get() : q;
							let res = yield Connection.query(rq);
							return res.rows.map(function (row) {
								return row.id;
							});
						} else if (opts.return == 'data') {
							let rq = q instanceof QueryBuilder ? q.get() : q;
							let res = yield Connection.query(rq);
							return res.rows;
						} else {
							return yield this.runQuery(q);
						}
					}, {
						loose: true
					})
				},
				runQuery: {
					method: CL.fun(function* (q) {
						let _private = this['private:headland.postgres.Row'];

						let rq = q instanceof QueryBuilder ? q.get() : q;
						let res = yield Connection.query(rq);

						let items = [];
						for (let i = 0; i < res.rows.length; i++) {
							let row = res.rows[i];
							let item = this.getObject(row.id);
							if (!item) item = new this(row, true);
							items.push(item);
						}

						return yield CL.wait(items);
					}, {
						loose: true
					})
				},
				filterFilters: {
					method: function (filters) {
						let _private = this['private:headland.postgres.Row'];

						let good = [];
						let bad = [];
						let have = {
							'null': true,
							'!null': true,
							'contains': true
						};
						for (let filter of filters) {
							let isarr = _2.isArray(filter);
							if (isarr && (ops[filter[1]] || have[filter[1]]) || !isarr && typeof filter == 'object') good.push(filter);else bad.push(filter);
						}
						return { good: good, bad: bad };
					}
				},
				delete: {
					method: CL.fun(function* (filter) {
						let _private = this['private:headland.postgres.Row'];

						if (filter) {
							let q = r.table(this.table);
							if (filter != 'everything') q = q.filter(filter);
							yield q.delete().run(get().connection, CL.cb());
						} else {
							CL.throw('HLPostgreSQLDelete', 'Filter or a string "everything" is required');
						}
					}, {
						loose: true
					})
				},
				configure: {
					method: CL.fun(function* () {
						let _private = this['private:headland.postgres.Row'];

						let res = yield Connection.query(`SELECT * FROM information_schema.columns WHERE table_name = $1 AND table_schema = 'public'`, [this.table]);
						let q;
						let num = 0;
						if (!res.rows.length) {
							let fields = [];
							_2.each(this.classInfo.fields, function (field, key) {
								if (key.match(/_id$/)) return;
								let add = function (key) {
									let str = fkey(key) + ' ' + (field.type() || 'varchar');
									if (field.primary) str += ' PRIMARY KEY';
									if (field.unique) str += ' UNIQUE';
									fields.push(str);
									num++;
								};
								if (field.type || field.sub) {
									add(key);
									if (field.translate) {
										for (let lang of Translation.languages) {
											add(key + '_' + lang);
										}
									}
								}
							});
							q = 'CREATE TABLE ' + fkey(this.table) + ' (' + fields.join(', ') + ')';
						} else {
							let current = {};
							_2.each(res.rows, function (row) {
								current[row.column_name] = row;
							});
							let actions = [];
							_2.each(this.classInfo.fields, function (field, key) {
								if (key.match(/_id$/)) return;
								let add = function (key) {
									let cur = current[key];
									if (!cur) {
										let str = 'ADD ' + fkey(key) + ' ' + (field.type() || 'varchar');
										if (field.primary) str += ' PRIMARY KEY';
										if (field.unique) str += ' UNIQUE';
										actions.push(str);
										num++;
									}
								};
								if (field.type || field.sub) {
									add(key);
									if (field.translate) {
										for (let lang of Translation.languages) {
											add(key + '_' + lang);
										}
									}
								}
							});
							q = 'ALTER TABLE ' + fkey(this.table) + ' ' + actions.join(', ');
						}
						if (num) {
							console.log(q);
							yield Connection.query(q);
						}
					}, {
						loose: true
					})
				}
			},
			privateStatics: {}
		});
		$$module.export(Row);
		$$module.loadDeferred(function* () {
			_2 = yield HL.require('lodash', false, true);
		});
	}), {
		immediate: [['headland', 'Model']],
		deferred: ['lodash']
	});
});
//# sourceMappingURL=/@/es6/headland.postgres.Row.js.map
